#include <math.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

#include <time.h>

#include <SDL.h>

#include <cairo.h>
#include <cairo-ft.h>

#include <ft2build.h>
#include FT_FREETYPE_H

void die(const char* message, ...) {
  va_list ap;
  va_start(ap, message);
  vfprintf(stderr, message, ap);
  va_end(ap);
  fputc('\n', stderr);
  exit(1);
};

struct timespec gettime() {
  struct timespec result;
  if (clock_gettime(CLOCK_MONOTONIC, &result) != 0)
    die("clock_gettime failed: %s", strerror(errno));
  return result;
};

struct timespec time_sub(const struct timespec* a, const struct timespec* b) {
  struct timespec result;
  result.tv_sec  = a->tv_sec  - b->tv_sec;
  result.tv_nsec = a->tv_nsec - b->tv_nsec;

  const long second = 1000000000; // ns
  if (result.tv_sec < 0) {
    --result.tv_sec;
    result.tv_nsec += second;
  } else if (result.tv_nsec > second) {
    ++result.tv_sec;
    result.tv_nsec -= second;
  };

  return result;
};

void nsleep(struct timespec* delay) {
  while (nanosleep(delay, delay) != 0)
    switch (errno) {
      case EFAULT:
        die("nanosleep failed to copy information from user space");
      case EINTR:
        continue;
      case EINVAL:
        die(
            "nanosleep: invalid value in the tv_nsec field: %ld", delay->tv_nsec
        );
    };
};

const int cell_size = 107;

void cairo_set_source_color(cairo_t* cairo, unsigned color) {
  cairo_set_source_rgb(
      cairo,
      (double)((color & 0xff0000) >> 16) / 0xff,
      (double)((color & 0x00ff00) >> 8)  / 0xff,
      (double)(color & 0x0000ff) / 0xff
  );
};

SDL_Texture* make_cells(SDL_Renderer* renderer) {
  const int gap = 15;

  const unsigned background[] = {
    0x776e65, // background
    0xeee4da, // 2
    0xede0c8, // 4
    0xf2b179, // 8
    0xf59563, // 16
    0xf67c5f, // 32
    0xf65e3b, // 64
    0xedcf72, // 128
    0xedcc61, // 256
    0xedc850, // 512
    0xedc53f, // 1024
    0xedc22e, // 2048
    0x3c3a32
  };
  const unsigned nbackgrounds = sizeof(background) / sizeof(*background);

  const unsigned foreground[] = { 0x776e65, 0xf9f6f2 };

  const unsigned font_size[] = { 55, 50, 45, 35, 30, 25, 22, 20, 18, 16, 14, 13, 12, 11, 10, 9 };

  FT_Library ft_lib;
  if (FT_Init_FreeType(&ft_lib) != 0) die("Failed to initialize FreeType2");
  FT_Face ft_face;
  const char* font_file = "ClearSans-Bold-webfont.ttf";
  switch (FT_New_Face(ft_lib, font_file, 0, &ft_face)) {
    case 0:
      break;
    case FT_Err_Unknown_File_Format:
      die("Unsupported font format in %s", font_file);
    default:
      die("Failed to open %s", font_file);
  };

  cairo_surface_t* csurface = cairo_image_surface_create(
      CAIRO_FORMAT_ARGB32, cell_size * 63, cell_size
  );

  cairo_t* c = cairo_create(csurface);

  cairo_font_face_t* cft_face = cairo_ft_font_face_create_for_ft_face(
      ft_face, 0
  );
  cairo_set_font_face(c, cft_face);

  cairo_set_line_width(c, 1);

  cairo_set_source_color(c, background[0]);
  cairo_rectangle(c, 0, 0, cell_size * 63, cell_size);
  cairo_fill(c);

  cairo_font_extents_t fe;
  cairo_font_extents(c, &fe);

  int n = 0;
  char text[21];
  for (int n = 1; n <= 63; ++n) {
#define get(array, n) \
    array[ \
      n < sizeof(array) / sizeof(*array) \
      ? n \
      : sizeof(array) / sizeof(*array) - 1 \
    ]
    cairo_set_source_color(c, get(background, n));
    cairo_rectangle(
        c,
        (n-1) * cell_size + gap / 2,
        0,
        cell_size - gap,
        cell_size - gap
    );
    cairo_fill(c);

    cairo_set_source_color(c, foreground[n < 3 ? 0 : 1]);
    int l = snprintf(text, 21, "%lu", 1UL << n);
    cairo_set_font_size(c, get(font_size, l - 1));
    cairo_text_extents_t te;
    cairo_text_extents(c, text, &te);
    cairo_move_to(
        c,
        0.5 * (cell_size - te.width ) - te.x_bearing + (n-1) * cell_size,
        0.5 * (cell_size - te.height) - te.y_bearing
    );
    cairo_show_text(c, text);
#undef get
  };

  cairo_surface_flush(csurface);

  cairo_status_t cstatus = cairo_surface_status(csurface);
  if (cstatus != CAIRO_STATUS_SUCCESS)
    die("Cairo: %s", cairo_status_to_string(cstatus));

  SDL_Surface* ssurface = SDL_CreateRGBSurfaceFrom(
      cairo_image_surface_get_data(csurface),
      cell_size * 63,
      cell_size,
      32,
      4 * 63 * cell_size,
      0x00ff0000,
      0x0000ff00,
      0x000000ff,
      0xff000000
  );
  if (!ssurface) die("Failed to create SDL surface: %s", SDL_GetError());

  SDL_Texture* cells = SDL_CreateTextureFromSurface(renderer, ssurface);
  if (!cells) die("Failed to create SDL texture: %s", SDL_GetError());

  SDL_FreeSurface(ssurface);
  cairo_font_face_destroy(cft_face);
  cairo_destroy(c);
  cairo_surface_destroy(csurface);
  FT_Done_Face(ft_face);
  FT_Done_FreeType(ft_lib);

  return cells;
};

int main() {
  SDL_Window* window = SDL_CreateWindow(
      "2048",
      SDL_WINDOWPOS_UNDEFINED,
      SDL_WINDOWPOS_UNDEFINED,
      8 * cell_size,
      8 * cell_size,
//      800,
//      600,
      0
  );
  if (!window) die("Failed to create SDL window: %s", SDL_GetError());

  SDL_Renderer* renderer = SDL_CreateRenderer(window, -1, SDL_RENDERER_ACCELERATED);
  if (!renderer) die("Failed to create SDL renderer: %s", SDL_GetError());

  SDL_Texture* cells = make_cells(renderer);

  struct timespec length;
  length.tv_sec  = 0;
  length.tv_nsec = 1000000000 / 60;
  while (1) {
    struct timespec start = gettime();

    SDL_Event event;
    while (SDL_PollEvent(&event))
      switch (event.type) {
        case SDL_QUIT:
          return 0;
      };

    if (SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255) != 0)
      die("Failed to set SDL render color: %s", SDL_GetError());
    if (SDL_RenderClear(renderer) != 0)
      die("Failed to clear window: %s", SDL_GetError());

    SDL_Rect src;
    src.y = 0;
    src.w = cell_size;
    src.h = cell_size;

    SDL_Rect dst;
    dst.w = cell_size;
    dst.h = cell_size;

    int n = 0;
    for (int x = 0; x < 8; ++x)
      for (int y = 0; y < 8; ++y)
        if (n < 64) {
          src.x = n++ * cell_size;
          dst.x = x * cell_size;
          dst.y = y * cell_size;
          if (SDL_RenderCopy(renderer, cells, &src, &dst) != 0)
            die("SDL_RenderCopy failed: %s", SDL_GetError());
        };

    SDL_RenderPresent(renderer);

    struct timespec end = gettime();
    struct timespec diff = time_sub(&end, &start);
    diff = time_sub(&length, &diff);
    if (diff.tv_sec >= 0 && diff.tv_nsec >= 0) nsleep(&diff);
  };

  return 0;
};
